/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.xwiki.lesscss.internal.compiler;

import java.io.File;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.inject.Singleton;

import org.apache.commons.lang.StringUtils;
import org.xwiki.component.annotation.Component;
import org.xwiki.lesscss.compiler.LESSCompilerException;
import org.xwiki.model.reference.DocumentReference;
import org.xwiki.model.reference.DocumentReferenceResolver;
import org.xwiki.model.reference.WikiReference;
import org.xwiki.wiki.descriptor.WikiDescriptorManager;

import com.xpn.xwiki.XWiki;
import com.xpn.xwiki.XWikiContext;
import com.xpn.xwiki.XWikiException;
import com.xpn.xwiki.doc.XWikiDocument;
import com.xpn.xwiki.objects.BaseObject;

/**
 * Component to get the directory of the base skin of any skin.
 *
 * @since 6.4M2
 * @version $Id: 1ead98babe6dfa047ec6a9ed09d8486dbd5b51f3 $
 */
@Component(roles = SkinDirectoryGetter.class)
@Singleton
public class SkinDirectoryGetter
{
    @Inject
    private Provider<XWikiContext> xcontextProvider;

    @Inject
    private WikiDescriptorManager wikiDescriptorManager;

    @Inject
    private DocumentReferenceResolver<String> documentReferenceResolver;

    /**
     * @param skin name of the skin
     * @return the directory of the base skin of the specified skin. The directory is relative to the webapp one.
     * @throws LESSCompilerException if problem occurs
     */
    public String getSkinDirectory(String skin) throws LESSCompilerException
    {
        // Is the skin a Wiki Document?
        return "/skins/" + getSkinDirectory(skin, new ArrayList<String>());
    }

    private String getSkinDirectory(String skin, List<String> alreadyVisitedSkins) throws LESSCompilerException
    {
        // Avoid infinite loop
        if (alreadyVisitedSkins.contains(skin)) {
            throw new LESSCompilerException(String.format("Infinite loop of 'baseskin' dependencies [%s].",
                    alreadyVisitedSkins.toString()), null);
        }
        alreadyVisitedSkins.add(skin);

        // Get the xwiki objects
        XWikiContext xcontext = xcontextProvider.get();
        XWiki xwiki = xcontext.getWiki();

        // Is the skin a Wiki Document?
        String currentWikiId = wikiDescriptorManager.getCurrentWikiId();
        DocumentReference skinDocRef = documentReferenceResolver.resolve(skin, new WikiReference(currentWikiId));
        if (skinDocRef != null && xwiki.exists(skinDocRef, xcontext)) {
            // Skin class
            DocumentReference skinClass = new DocumentReference(skinDocRef.getWikiReference().getName(),
                    "XWiki", "XWikiSkins");
            // Verify that the document is a skin by checking if a skin object is attached or not
            try {
                XWikiDocument skinDoc = xwiki.getDocument(skinDocRef, xcontext);
                BaseObject skinObj = skinDoc.getXObject(skinClass);
                if (skinObj != null) {
                    // Get the "baseskin" property of the skin
                    String baseSkin = skinObj.getStringValue("baseskin");
                    if (StringUtils.isBlank(baseSkin)) {
                        throw new LESSCompilerException(
                                String.format("Failed to get the base skin of the skin [%s].", skin),
                                null);
                    }
                    // Recursively get the skin directory from the baseskin
                    return getSkinDirectory(baseSkin, alreadyVisitedSkins);
                }
            } catch (XWikiException e) {
                throw new LESSCompilerException(String.format("Failed to get the document [%s].", skinDocRef), e);
            }
        }
        // If not, we assume it is a skin on the filesystem
        return skin;
    }

    /**
     * Get the full path of the directory where LESS skin file resources are stored inside a skin.
     * @param skin the name of the skin
     * @return the directory where LESS skin file resources are stored inside a skin.
     * @throws LESSCompilerException if problem occurs
     */
    public Path getLESSSkinFilesDirectory(String skin) throws LESSCompilerException
    {
        // Get the XWiki object
        XWikiContext xcontext = xcontextProvider.get();
        XWiki xwiki = xcontext.getWiki();

        // Get the skin directory, where LESS resources are located
        String path = this.getSkinDirectory(skin) + "/less";
        String realPath = xwiki.getEngineContext().getRealPath(path);
        File lessDirectory = new File(realPath);
        if (!lessDirectory.exists() || !lessDirectory.isDirectory()) {
            throw new LESSCompilerException(
                    String.format("The path [%s] is not a directory or does not exists.",
                            path));
        }

        return Paths.get(realPath);
    }
}
